from gomi.ankiconnect import invoke


def list_decks():
    decks = invoke("deckNames")
    print("List of decks:")
    for deck in decks:
        print(deck)


if __name__ == "__main__":
    list_decks()



# Importer imports note types from ../templates/ to Anki.
# Copyright: Ren Tatsumoto <tatsu at autistici.org>
# License: GNU GPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import json
import os
from typing import Any, Collection, Iterable

from .ankiconnect import invoke, request_model_names
from .common import CardTemplate, NoteType, get_used_fonts, select
from .consts import NOTE_TYPES_DIR, CSS_FILENAME, FRONT_FILENAME, BACK_FILENAME, JSON_FILENAME, FONTS_DIR


def read_css(model_dir_name: str) -> str:
    with open(NOTE_TYPES_DIR / model_dir_name / CSS_FILENAME, encoding="utf8") as f:
        return f.read()


def read_card_templates(model_dir_name: str, template_names: list[str]) -> list[CardTemplate]:
    templates = []
    for template_name in template_names:
        dir_path = NOTE_TYPES_DIR / model_dir_name / template_name
        with (
            open(dir_path / FRONT_FILENAME, encoding="utf8") as front,
            open(dir_path / BACK_FILENAME, encoding="utf8") as back,
        ):
            templates.append(CardTemplate(template_name, front.read(), back.read()))
    return templates


def read_model_dict(model_dir_name: str) -> dict[str, Any]:
    with open(os.path.join(NOTE_TYPES_DIR, model_dir_name, JSON_FILENAME), encoding="utf8") as f:
        return json.load(f)


def read_model(model_dir_name: str) -> NoteType:
    model_dict = read_model_dict(model_dir_name)
    return NoteType(
        name=model_dict["modelName"],
        fields=model_dict["inOrderFields"],
        css=read_css(model_dir_name),
        templates=read_card_templates(model_dir_name, model_dict["cardTemplates"]),
    )


def format_import(model: NoteType) -> dict[str, Any]:
    return {
        "modelName": model.name,
        "inOrderFields": model.fields,
        "css": model.css,
        "cardTemplates": [
            {
                "Name": template.name,
                "Front": template.front,
                "Back": template.back,
            }
            for template in model.templates
        ],
    }


def send_note_type(model: NoteType):
    template_json = format_import(model)
    while template_json["modelName"] in request_model_names():
        template_json["modelName"] = input("Model with this name already exists. Enter new name: ")
    invoke("createModel", **template_json)


def available_fonts(required_fonts: Collection[str]) -> Iterable[str]:
    """Filter required fonts and leave only those available on disk."""
    for file in os.listdir(FONTS_DIR):
        if file in required_fonts:
            yield file


def store_fonts(fonts: list[str]):
    for file in available_fonts(fonts):
        invoke("storeMediaFile", filename=file, path=os.path.join(FONTS_DIR, file))


def import_note_type():
    if model_dir_name := select(os.listdir(NOTE_TYPES_DIR)):
        print(f"Selected model: {model_dir_name}")
        model = read_model(model_dir_name)
        store_fonts(get_used_fonts(model.css))
        send_note_type(model)
        print("Done.")



# Overwriter acts like Updater but allows mapping models with different names.
# Copyright: Ren Tatsumoto <tatsu at autistici.org>
# License: GNU GPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
import os

from .ankiconnect import request_model_names
from .common import get_used_fonts, select
from .consts import NOTE_TYPES_DIR
from .importer import read_model, store_fonts
from .updater import send_note_type


def overwrite_note_type():
    anki_models = request_model_names()
    models_on_disk = {(model := read_model(dir_name)).name: model for dir_name in os.listdir(NOTE_TYPES_DIR)}
    model_name_on_disk = select(list(models_on_disk), "Take stored model: ")
    model_name_in_anki = select(anki_models, "Replace templates in model: ")

    print(f"Writing templates from {model_name_on_disk} onto {model_name_in_anki}...")

    model = models_on_disk[model_name_on_disk]
    store_fonts(get_used_fonts(model.css))
    send_note_type(model.rename(model_name_in_anki))

    print("Done.")



# Updater updates previously imported note types.
# Copyright: Ren Tatsumoto <tatsu at autistici.org>
# License: GNU GPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
import os
from typing import Any

from .ankiconnect import invoke, request_model_names
from .common import NoteType, get_used_fonts, select
from .consts import NOTE_TYPES_DIR
from .importer import read_model, store_fonts


def format_templates(model: NoteType) -> dict[str, Any]:
    return {
        "model": {
            "name": model.name,
            "templates": {
                template.name: {"Front": template.front, "Back": template.back} for template in model.templates
            },
        }
    }


def format_styling(model: NoteType) -> dict[str, Any]:
    return {"model": {"name": model.name, "css": model.css}}


def send_note_type(model: NoteType):
    invoke("updateModelTemplates", **format_templates(model))
    invoke("updateModelStyling", **format_styling(model))


def update_note_type():
    anki_models = request_model_names()
    models_on_disk = {(model := read_model(dir_name)).name: model for dir_name in os.listdir(NOTE_TYPES_DIR)}
    updatable_models = [model_name for model_name in models_on_disk if model_name in anki_models]
    if not updatable_models:
        print("No note types can be updated.")
        return
    if model_name := select(updatable_models):
        print(f"Selected note type: {model_name}")
        model = models_on_disk[model_name]
        store_fonts(get_used_fonts(model.css))
        send_note_type(model)
        print("Done.")



# Exporter exports note types from Anki to ../templates/
# Copyright: Ren Tatsumoto <tatsu at autistici.org>
# License: GNU GPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import base64
import json
import os
import pathlib
import random
import shutil
from os import DirEntry
from typing import Any

from .ankiconnect import invoke, request_model_names
from .common import CardTemplate, NoteType, get_used_fonts, select
from .consts import (
    NOTE_TYPES_DIR,
    FRONT_FILENAME,
    BACK_FILENAME,
    JSON_FILENAME,
    CSS_FILENAME,
    README_FILENAME,
    JSON_INDENT,
    FONTS_DIR,
)


def fetch_card_templates(model_name: str) -> list[CardTemplate]:
    return [
        CardTemplate(name, val["Front"], val["Back"])
        for name, val in invoke("modelTemplates", modelName=model_name).items()
    ]


def fetch_template(model_name: str) -> NoteType:
    return NoteType(
        name=model_name,
        fields=invoke("modelFieldNames", modelName=model_name),
        css=invoke("modelStyling", modelName=model_name)["css"],
        templates=fetch_card_templates(model_name),
    )


def select_model_dir_path(model_name: str) -> pathlib.Path:
    dir_path = NOTE_TYPES_DIR / model_name
    dir_content = frozenset(os.listdir(NOTE_TYPES_DIR))

    if model_name in dir_content:
        ans = input("Template with this name already exists. Overwrite [y/N]? ")
        if ans.lower() != "y":
            while dir_path.name in dir_content:
                dir_path = NOTE_TYPES_DIR / f"{model_name}_{random.randint(0, 9999)}"

    return dir_path


def write_card_templates(model_dir_path: pathlib.Path, templates: list[CardTemplate]) -> None:
    for template in templates:
        dir_path = os.path.join(model_dir_path, template.name)
        if not os.path.isdir(dir_path):
            os.mkdir(dir_path)
        for filename, content in zip((FRONT_FILENAME, BACK_FILENAME), (template.front, template.back)):
            with open(os.path.join(dir_path, filename), "w", encoding="utf8") as f:
                f.write(content)


def format_export(model: NoteType) -> dict[str, Any]:
    return {
        "modelName": model.name,
        "inOrderFields": model.fields,
        "cardTemplates": [template.name for template in model.templates],
    }


def remove_deleted_templates(model_dir_path: pathlib.Path, templates: list[str]) -> None:
    entry: DirEntry
    for entry in os.scandir(model_dir_path):
        if entry.is_dir() and entry.name not in templates:
            shutil.rmtree(entry.path)


def save_note_type(model: NoteType):
    dir_path = select_model_dir_path(model.name)
    json_path = dir_path / JSON_FILENAME
    css_path = dir_path / CSS_FILENAME
    readme_path = dir_path / README_FILENAME

    if not os.path.isdir(dir_path):
        os.mkdir(dir_path)

    with open(json_path, "w", encoding="utf8") as f:
        json.dump(format_export(model), f, indent=JSON_INDENT, ensure_ascii=False)

    with open(css_path, "w", encoding="utf8") as f:
        f.write(model.css)

    write_card_templates(dir_path, model.templates)
    remove_deleted_templates(dir_path, [template.name for template in model.templates])

    if not readme_path.is_file():
        with open(readme_path, "w", encoding="utf8") as f:
            f.write(f"# {model.name}\n\n*Description and screenshots here.*")


def save_fonts(model: NoteType) -> None:
    linked_fonts = get_used_fonts(model.css)
    for font in linked_fonts:
        if file_b64 := invoke("retrieveMediaFile", filename=font):
            with open(os.path.join(FONTS_DIR, font), "bw") as f:
                f.write(base64.b64decode(file_b64))


def export_note_type():
    if model := select(request_model_names()):
        print(f"Selected model: {model}")
        template = fetch_template(model)
        save_fonts(template)
        save_note_type(template)
        print("Done.")



# Taken from https://github.com/FooSoft/anki-connect
# Copyright: Ren Tatsumoto <tatsu at autistici.org>
# License: GNU GPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

__all__ = ["invoke", "request_model_names"]

import json
import urllib.request

from .common import ANTPError


def request(action, **params):
    return {"action": action, "params": params, "version": 6}


def invoke(action, **params):
    request_json = json.dumps(request(action, **params)).encode("utf-8")
    response = json.load(
        urllib.request.urlopen(urllib.request.Request("http://127.0.0.1:8765", request_json), timeout=10)
    )
    if len(response) != 2:
        raise ANTPError("response has an unexpected number of fields")
    if "error" not in response:
        raise ANTPError("response is missing required error field")
    if "result" not in response:
        raise ANTPError("response is missing required result field")
    if response["error"] is not None:
        raise ANTPError(response["error"])
    return response["result"]


def request_model_names() -> list[str]:
    return invoke("modelNames")



# Copyright: Ren Tatsumoto <tatsu at autistici.org>
# License: GNU GPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import os
import pathlib
import sys
from typing import Callable
from urllib.error import URLError

from .common import ANTPError, init
from .consts import NOTE_TYPES_DIR
from .exporter import export_note_type
from .importer import import_note_type
from .overwriter import overwrite_note_type
from .updater import update_note_type


def wrap_errors(fn: Callable):
    try:
        fn()
    except URLError:
        print("Couldn't connect. Make sure Anki is open and AnkiConnect is installed.")
    except ANTPError as ex:
        print(ex)
    except KeyboardInterrupt:
        print("\nAborted.")


def program_name() -> str:
    return os.path.basename(sys.argv[0])


def print_help():
    options = (
        ("import", "Add one of the stored note types to Anki."),
        ("update", "Overwrite a previously imported note type with new data. " "Fields will not be updated."),
        (
            "overwrite",
            "Overwrite a note type in Anki with new data from a stored note type. " "Fields will not be updated.",
        ),
        ("export", "Save your note type to disk as a template."),
        ("list", "List models stored in the templates folder."),
        ("-v, --verbose", "Show detailed info when errors occur."),
    )
    print(f"Usage: {program_name()} [OPTIONS]\n\n" "Options:")
    col_width = [max(len(word) for word in col) + 2 for col in zip(*options)]
    for row in options:
        print(" " * 4, "".join(col.ljust(col_width[i]) for i, col in enumerate(row)), sep="")


def list_stored_note_types():
    print("\n".join(os.listdir(NOTE_TYPES_DIR)))


def is_correct_cwd():
    return pathlib.Path.cwd().joinpath(".git").is_dir()


def main() -> int:
    if not is_correct_cwd():
        print("Current directory is not a git repository. Run `git init`.")
        return 1

    if len(sys.argv) < 2:
        print("No action provided.")
        print_help()
        return 1

    init()

    action = None
    wrap = True

    for arg in sys.argv[1:]:
        match arg:
            case "export":
                action = export_note_type
            case "import":
                action = import_note_type
            case "update":
                action = update_note_type
            case "overwrite":
                action = overwrite_note_type
            case "list":
                action = list_stored_note_types
            case "-v" | "--verbose":
                wrap = False

    if action and wrap:
        wrap_errors(action)
    elif action:
        action()
    else:
        print("Unknown action.")
        print_help()
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright: Ren Tatsumoto <tatsu at autistici.org>
# License: GNU GPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import dataclasses
import re
from dataclasses import dataclass
from .consts import FONTS_DIR, NOTE_TYPES_DIR


class ANTPError(Exception):
    pass


@dataclass(frozen=True)
class CardTemplate:
    name: str
    front: str
    back: str


@dataclass(frozen=True)
class NoteType:
    name: str
    fields: list[str]
    css: str
    templates: list[CardTemplate]

    def rename(self, new_name: str):
        return dataclasses.replace(self, name=new_name)


def read_num(msg: str = "Input number: ", min_val: int = 0, max_val: int | None = None) -> int:
    try:
        resp = int(input(msg))
    except ValueError as ex:
        raise ANTPError(ex) from ex
    if resp < min_val or (max_val and resp > max_val):
        raise ANTPError("Value out of range.")
    return resp


def select(items: list[str], msg: str = "Select item number: ") -> str | None:
    if not items:
        print("Nothing to show.")
        return None

    for idx, model in enumerate(items):
        print(f"{idx}: {model}")
    print()

    idx = read_num(msg, max_val=len(items) - 1)
    return items[idx]


def get_used_fonts(template_css: str):
    return re.findall(r"url\([\"'](\w+\.(?:[ot]tf|woff\d?))[\"']\)", template_css, re.IGNORECASE)


def init():
    for path in (NOTE_TYPES_DIR, FONTS_DIR):
        path.mkdir(exist_ok=True)



# Copyright: Ren Tatsumoto <tatsu at autistici.org>
# License: GNU GPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import pathlib

JSON_INDENT = 4
JSON_FILENAME = "template.json"
CSS_FILENAME = "template.css"
FRONT_FILENAME = "front.html"
BACK_FILENAME = "back.html"
README_FILENAME = "README.md"
THIS_DIR = pathlib.Path.cwd()
NOTE_TYPES_DIR = THIS_DIR / "templates"
FONTS_DIR = THIS_DIR / "fonts"
